@@ -2,23 +2,15 @@
     Classes and methods to maintain any bibtex information that is stored
     outside the doctree.
 
-    .. autoclass:: BibliographyKey
-        :members:
-
-    .. autoclass:: BibliographyValue
-        :members:
-
     .. autoclass:: Citation
         :members:
 
-    .. autoclass:: CitationRef
-        :members:
-
     .. autoclass:: BibtexDomain
         :members:
 """
 
 import ast
+from typing import TYPE_CHECKING
 from typing import List, Dict, NamedTuple, cast, Iterable, Tuple, Set
 
 import docutils.frontend
@@ -28,19 +20,25 @@ import docutils.utils
 import sphinx.util
 import re
 
-from pybtex.database import Entry
 from pybtex.plugin import find_plugin
-import pybtex.style.formatting
-from pybtex.style import FormattedEntry
-from sphinx.addnodes import pending_xref
-from sphinx.builders import Builder
 from sphinx.domains import Domain
-from sphinx.environment import BuildEnvironment
 from sphinx.errors import ExtensionError
 from sphinx.util.nodes import make_refnode
 
 from .bibfile import BibFile, normpath_filename, process_bibfile
 
+if TYPE_CHECKING:
+    from pybtex.database import Entry
+    from pybtex.style import FormattedEntry
+    from pybtex.style.formatting import BaseStyle
+    from sphinx.addnodes import pending_xref
+    from sphinx.application import Sphinx
+    from sphinx.builders import Builder
+    from sphinx.environment import BuildEnvironment
+    from .directives import BibliographyKey, BibliographyValue
+    from .roles import CitationRef
+
+
 logger = sphinx.util.logging.getLogger(__name__)
 
 
@@ -192,40 +190,18 @@ def get_docnames(env):
         docname = nextdoc
 
 
-class BibliographyKey(NamedTuple):
-    docname: str
-    id_: str
-
-
-class BibliographyValue(NamedTuple):
-    """Contains information about a bibliography directive."""
-    line: int            #: Line number of the directive in the document.
-    bibfiles: List[str]  #: List of bib files for this directive.
-    style: str           #: The pybtex style.
-    list_: str           #: The list type.
-    enumtype: str        #: The sequence type (for enumerated lists).
-    start: int           #: The start of the sequence (for enumerated lists).
-    labelprefix: str     #: String prefix for pybtex generated labels.
-    keyprefix: str       #: String prefix for citation keys.
-    filter_: ast.AST     #: Parsed filter expression.
-    citation_nodes: Dict[str, docutils.nodes.citation]  #: key -> citation node
-
-
 class Citation(NamedTuple):
     """Information about a citation."""
-    citation_id: str                   #: Unique id of this citation.
-    bibliography_key: BibliographyKey  #: Key of its bibliography directive.
-    key: str                           #: Key (with prefix).
-    label: str                         #: Label (with prefix).
-    formatted_entry: FormattedEntry    #: Entry as formatted by pybtex.
+    citation_id: str                     #: Unique id of this citation.
+    bibliography_key: "BibliographyKey"  #: Key of its bibliography directive.
+    key: str                             #: Key (with prefix).
+    label: str                           #: Label (with prefix).
+    formatted_entry: "FormattedEntry"    #: Entry as formatted by pybtex.
 
 
-class CitationRef(NamedTuple):
-    """Information about a citation reference."""
-    citation_ref_id: str  #: Unique id of this citation reference.
-    docname: str          #: Document name.
-    line: int             #: Line number.
-    keys: List[str]       #: Citation keys (including key prefix).
+def env_updated(app: "Sphinx", env: "BuildEnvironment") -> Iterable[str]:
+    domain = cast(BibtexDomain, env.get_domain('cite'))
+    return domain.env_updated()
 
 
 class BibtexDomain(Domain):
@@ -233,32 +209,50 @@ class BibtexDomain(Domain):
 
     name = 'cite'
     label = 'BibTeX Citations'
-    data_version = 2
+    data_version = 3
+    initial_data = dict(
+        bibfiles={},
+        bibliography_header=docutils.nodes.paragraph(),
+        footbibliography_header=docutils.nodes.paragraph(),
+        bibliographies={},
+        citations=[],
+        citation_refs=[],
+    )
 
     @property
     def bibfiles(self) -> Dict[str, BibFile]:
         """Map each bib filename to some information about the file (including
         the parsed data).
         """
-        return self.data.setdefault('bibfiles', {})
+        return self.data['bibfiles']
 
     @property
-    def bibliographies(self) -> Dict[BibliographyKey, BibliographyValue]:
+    def bibliography_header(self) -> docutils.nodes.Element:
+        return self.data['bibliography_header']
+
+    @property
+    def footbibliography_header(self) -> docutils.nodes.Element:
+        return self.data['footbibliography_header']
+
+    @property
+    def bibliographies(self) -> Dict["BibliographyKey", "BibliographyValue"]:
         """Map storing information about each bibliography directive."""
-        return self.data.setdefault('bibliographies', {})
+        return self.data['bibliographies']
 
     @property
     def citations(self) -> List[Citation]:
         """Citation data."""
-        return self.data.setdefault('citations', [])
+        return self.data['citations']
 
     @property
-    def citation_refs(self) -> List[CitationRef]:
+    def citation_refs(self) -> List["CitationRef"]:
         """Citation reference data."""
-        return self.data.setdefault('citation_refs', [])
+        return self.data['citation_refs']
 
-    def __init__(self, env: BuildEnvironment):
+    def __init__(self, env: "BuildEnvironment"):
         super().__init__(env)
+        # connect env-updated
+        env.app.connect('env-updated', env_updated)
         # check config
         if env.app.config.bibtex_bibfiles is None:
             raise ExtensionError(
@@ -270,26 +264,25 @@ class BibtexDomain(Domain):
                 env.app.config.bibtex_encoding)
         # parse bibliography headers
         for directive in ("bibliography", "footbibliography"):
-            conf_name = "bibtex_{0}_header".format(directive)
-            if not hasattr(env, conf_name):
+            header = getattr(env.app.config, "bibtex_%s_header" % directive)
+            if header:
                 parser = docutils.parsers.rst.Parser()
                 settings = docutils.frontend.OptionParser(
                     components=(docutils.parsers.rst.Parser,)
                 ).get_default_values()
                 document = docutils.utils.new_document(
-                    "{0}_header".format(directive), settings)
-                parser.parse(getattr(env.app.config, conf_name), document)
-                setattr(env, conf_name,
-                        document[0] if len(document) > 0 else None)
+                    "%s_header" % directive, settings)
+                parser.parse(header, document)
+                if len(document) > 0:
+                    self.data["%s_header" % directive] = document[0]
 
     def clear_doc(self, docname: str) -> None:
         self.data['citations'] = [
             citation for citation in self.citations
             if citation.bibliography_key.docname != docname]
         self.data['citation_refs'] = [
-            ref for ref in self.citation_refs
-            if ref.docname != docname]
-        for bib_key, bib_value in list(self.bibliographies.items()):
+            ref for ref in self.citation_refs if ref.docname != docname]
+        for bib_key in list(self.bibliographies.keys()):
             if bib_key.docname == docname:
                 del self.bibliographies[bib_key]
 
@@ -300,21 +293,20 @@ class BibtexDomain(Domain):
         for citation_ref in otherdata['citation_refs']:
             if citation_ref.docname in docnames:
                 self.citation_refs.append(citation_ref)
-        # citations created during check_consistency so never pickled
-        assert not self.citations
-        assert not otherdata['citations']
+        # 'citations' domain data calculated in env_updated
 
-    def check_consistency(self) -> None:
+    def env_updated(self) -> Iterable[str]:
         # This function is called when all doctrees are parsed,
         # but before any post transforms are applied. We use it to
         # determine which citations will be added to which bibliography
         # directive, and also to format the labels. We need to format
-        # the labels and construct the citation ids here because they must be
-        # known when resolve_xref is called.
+        # the labels here because they must be known when resolve_xref is
+        # called.
+        self.citations.clear()  # might have been restored from pickle
         docnames = list(get_docnames(self.env))
         # we keep track of this to quickly check for duplicates
         used_keys: Set[str] = set()
-        used_labels: Dict[str, Set[str]] = {}
+        used_labels: Dict[str, str] = {}
         for bibliography_key, bibliography in self.bibliographies.items():
             for formatted_entry in self.get_formatted_entries(
                     bibliography_key, docnames):
@@ -322,7 +314,7 @@ class BibtexDomain(Domain):
                 label = bibliography.labelprefix + formatted_entry.label
                 if bibliography.list_ == 'citation' and key in used_keys:
                     logger.warning(
-                        'duplicate citation for key %s' % key,
+                        'duplicate citation for key "%s"' % key,
                         location=(bibliography_key.docname, bibliography.line))
                 self.citations.append(Citation(
                     citation_id=bibliography.citation_nodes[key]['ids'][0],
@@ -333,23 +325,28 @@ class BibtexDomain(Domain):
                 ))
                 if bibliography.list_ == 'citation':
                     used_keys.add(key)
-                    used_labels.setdefault(label, set()).add(key)
-        for label, keys in used_labels.items():
-            if len(keys) > 1:
-                logger.warning(
-                    'duplicate label %s for keys %s' % (
-                        label, ','.join(sorted(keys))))
+                    if label not in used_labels:
+                        used_labels[label] = key
+                    elif used_labels[label] != key:
+                        # if used_label[label] == key then already
+                        # duplicate key warning
+                        logger.warning(
+                            'duplicate label "%s" for keys "%s" and "%s"' % (
+                                label, used_labels[label], key),
+                            location=(bibliography_key.docname,
+                                      bibliography.line))
+        return []  # expects list of updated docnames
 
-    def resolve_xref(self, env: BuildEnvironment, fromdocname: str,
-                     builder: Builder, typ: str, target: str,
-                     node: pending_xref, contnode: docutils.nodes.Element
+    def resolve_xref(self, env: "BuildEnvironment", fromdocname: str,
+                     builder: "Builder", typ: str, target: str,
+                     node: "pending_xref", contnode: docutils.nodes.Element
                      ) -> docutils.nodes.Element:
         """Replace node by list of citation references (one for each key)."""
         keys = [key.strip() for key in target.split(',')]
         if builder.name != 'latex':
-            node = docutils.nodes.inline(rawsource=target, text='[')
+            citations_node = docutils.nodes.inline(rawsource=target, text='[')
         else:
-            node = docutils.nodes.inline(rawsource=target, text='')
+            citations_node = docutils.nodes.inline(rawsource=target, text='')
         # map citation keys that can be resolved to their citation data
         citations = {
             cit.key: cit for cit in self.citations
@@ -360,8 +357,9 @@ class BibtexDomain(Domain):
                 citation = citations[key]
             except KeyError:
                 # TODO can handle missing reference warning using the domain
-                logger.warning('could not find bibtex key %s' % key)
-                node += docutils.nodes.inline('', key)
+                logger.warning('could not find bibtex key "%s"' % key,
+                               location=node)
+                citations_node += docutils.nodes.inline('', key)
                 continue
             refcontnode = docutils.nodes.inline('', citation.label)
             if builder.name == 'latex':
@@ -376,12 +374,12 @@ class BibtexDomain(Domain):
                     builder, fromdocname,
                     citation.bibliography_key.docname,
                     citation.citation_id, refcontnode)
-            node += refnode
+            citations_node += refnode
             if i != len(keys) - 1 and builder.name != 'latex':
-                node += docutils.nodes.Text(',')
+                citations_node += docutils.nodes.Text(',')
         if builder.name != 'latex':
-            node += docutils.nodes.Text(']')
-        return node
+            citations_node += docutils.nodes.Text(']')
+        return citations_node
 
     def get_all_cited_keys(self, docnames):
         """Yield all citation keys for given *docnames* in order, then
@@ -393,7 +391,7 @@ class BibtexDomain(Domain):
                 yield key
 
     def get_entries(
-            self, bibfiles: List[str]) -> Iterable[Entry]:
+            self, bibfiles: List[str]) -> Iterable["Entry"]:
         """Return all bibliography entries from the bib files, unsorted (i.e.
         in order of appearance in the bib files.
         """
@@ -402,8 +400,8 @@ class BibtexDomain(Domain):
                 yield entry
 
     def get_filtered_entries(
-            self, bibliography_key: BibliographyKey
-            ) -> Iterable[Tuple[str, Entry]]:
+            self, bibliography_key: "BibliographyKey"
+            ) -> Iterable[Tuple[str, "Entry"]]:
         """Return unsorted bibliography entries filtered by the filter
         expression.
         """
@@ -431,8 +429,8 @@ class BibtexDomain(Domain):
                 yield key, entry
 
     def get_sorted_entries(
-            self, bibliography_key: BibliographyKey, docnames: List[str]
-            ) -> Iterable[Tuple[str, Entry]]:
+            self, bibliography_key: "BibliographyKey", docnames: List[str]
+            ) -> Iterable[Tuple[str, "Entry"]]:
         """Return filtered bibliography entries sorted by citation order."""
         entries = dict(
             self.get_filtered_entries(bibliography_key))
@@ -448,15 +446,14 @@ class BibtexDomain(Domain):
             yield key, entry
 
     def get_formatted_entries(
-            self, bibliography_key: BibliographyKey, docnames: List[str]
-            ) -> Iterable[FormattedEntry]:
+            self, bibliography_key: "BibliographyKey", docnames: List[str]
+            ) -> Iterable["FormattedEntry"]:
         """Get sorted bibliography entries along with their pybtex labels,
         with additional sorting and formatting applied from the pybtex style.
         """
         bibliography = self.bibliographies[bibliography_key]
         entries = dict(
             self.get_sorted_entries(bibliography_key, docnames))
-        style = cast(
-            pybtex.style.formatting.BaseStyle,
-            find_plugin('pybtex.style.formatting', bibliography.style)())
+        style = cast("BaseStyle", find_plugin(
+            'pybtex.style.formatting', bibliography.style)())
         return style.format_entries(entries.values())
