@@ -9,3 +9,8 @@ collect_ignore = ['roots']
 @pytest.fixture(scope='session')
 def rootdir():
     return path(__file__).parent.abspath() / 'roots'
+
+
+# monkey patch for path class on old sphinx versions
+if not hasattr(path, "read_text"):
+    path.read_text = path.text
