@@ -11,7 +11,7 @@ import pytest
 
 @pytest.mark.sphinx('html', testroot='autodoc')
 def test_autodoc(app, warning):
-    app.builder.build_all()
+    app.build()
     assert not warning.getvalue()
     output = (app.outdir / "doc_cite.html").read_text()
     labels = ['One', 'Two', 'Thr', 'Fou', 'Fiv', 'Six', 'Sev', 'Eig', 'Nin',
@@ -36,3 +36,10 @@ def test_autodoc(app, warning):
         assert match
         id_ = match.group('id_')
         assert len(common.html_footnote_refs(refid=id_).findall(output2)) == 1
+
+
+# test that sphinx [source] links do not generate a warning (issue 17)
+@pytest.mark.sphinx('html', testroot='autodoc_viewcode')
+def test_autodoc_viewcode(app, warning):
+    app.build()
+    assert not warning.getvalue()
